"""
The MIT License (MIT)

Copyright (c) 2015-2017 Kim Blomqvist

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

import os
import encodings

import click
from click import ClickException

from . import yasha
from .tests import TESTS
from .filters import FILTERS
from .classes import CLASSES
from .parsers import PARSERS

def print_version(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return
    click.echo(yasha.__version__)
    ctx.exit()

def parse_variable_file(file):
    try:
        file_extension = os.path.splitext(file.name)[1]
        return PARSERS[file_extension](file)
    except AttributeError:
        return dict()
    except KeyError:
        error = "Unkown variable file extension '{}'"
        raise ClickException(error.format(file_extension))

def load_python_module(file):
    try:
        from importlib.machinery import SourceFileLoader
        loader = SourceFileLoader('yasha_extensions', file.name)
        module = loader.load_module()
    except ImportError:  # Fallback to Python2
        import imp
        desc = (".py", "rb", imp.PY_SOURCE)
        module = imp.load_module('yasha_extensions', file, file.name, desc)
    return module

def load_extensions(file):
    from jinja2.ext import Extension
    import inspect

    tests   = dict()
    filters = dict()
    parsers = dict()
    classes = []

    try:
        module = load_python_module(file)
    except NameError as e:
        msg = 'Unable to load extensions, {}'
        raise ClickException(msg.format(e))
    except SyntaxError as e:
        msg = "Unable to load extensions\n{} ({}, line {})"
        error = e.msg[0].upper() + e.msg[1:]
        filename = os.path.relpath(e.filename)
        raise ClickException(msg.format(error, filename, e.lineno))

    for attr in [getattr(module, x) for x in dir(module)]:
        if inspect.isfunction(attr):
            if attr.__name__.startswith('test_'):
                name = attr.__name__[5:]
                tests[name] = attr
            if attr.__name__.startswith('filter_'):
                name = attr.__name__[7:]
                filters[name] = attr
            if attr.__name__.startswith('parse_'):
                name = attr.__name__[6:]
                parsers['.' + name] = attr
        if inspect.isclass(attr):
            if issubclass(attr, Extension):
                classes.append(attr)

    try:
        TESTS.update(module.TESTS)
    except AttributeError:
        TESTS.update(tests)

    try:
        FILTERS.update(module.FILTERS)
    except AttributeError:
        FILTERS.update(filters)

    try:
        PARSERS.update(module.PARSERS)
    except AttributeError:
        PARSERS.update(parsers)


@click.command(context_settings=dict(
    help_option_names=["-h", "--help"],
    ignore_unknown_options=True,
))
@click.argument("template_variables", nargs=-1, type=click.UNPROCESSED)
@click.argument("template", type=click.File("rb"))
@click.option("--output", "-o", type=click.File("wb"), help="Place the rendered template into FILENAME.")
@click.option("--variables", "-v", type=click.File("rb"), help="Read template variables from FILENAME. Built-in parsers are JSON, YAML, TOML and XML.")
@click.option("--extensions", "-e", envvar='YASHA_EXTENSIONS', type=click.File("rb"), help="Read template extensions from FILENAME. A Python file is expected.")
@click.option("--encoding", "-c", default=yasha.ENCODING, help="Default is UTF-8.")
@click.option("--include_path", "-I", type=click.Path(exists=True, file_okay=False), multiple=True, help="Add DIRECTORY to the list of directories to be searched for the referenced templates.")
@click.option("--no-variable-file", is_flag=True, help="Omit template variable file.")
@click.option("--no-extension-file", is_flag=True, help="Omit template extension file.")
@click.option("--no-trim-blocks", is_flag=True, help="Load Jinja with trim_blocks=False.")
@click.option("--no-lstrip-blocks", is_flag=True, help="Load Jinja with lstrip_blocks=False.")
@click.option("--keep-trailing-newline", is_flag=True, help="Load Jinja with keep_trailing_newline=True.")
@click.option("-M", is_flag=True, help="Outputs Makefile compatible list of dependencies. Doesn't render the template.")
@click.option("-MD", is_flag=True, help="Creates Makefile compatible .d file alongside the rendered template.")
@click.option('--version', is_flag=True, callback=print_version, expose_value=False, is_eager=True, help="Print version and exit.")
def cli(template_variables, template, output, variables, extensions, encoding, include_path, no_variable_file, no_extension_file, no_trim_blocks, no_lstrip_blocks, keep_trailing_newline, m, md):
    """Reads the given Jinja TEMPLATE and renders its content
    into a new file. For example, a template called 'foo.c.j2'
    will be written into 'foo.c' in case the output file is not
    explicitly given.

    Template variables can be defined in a separate file or
    given as part of the command-line call, e.g.

        yasha --hello=world -o output.txt template.j2

    defines a variable 'hello' for a template like:
    
        Hello {{ hello }} !
    """

    # Set the encoding of the template file
    if encodings.search_function(encoding) is None:
        msg = "Unrecognized encoding name '{}'"
        raise ClickException(msg.format(encoding))
    yasha.ENCODING = encoding

    # Append include path of referenced templates
    include_path = [os.path.dirname(template.name)] + list(include_path)

    if not extensions or not variables:
        template_companion = yasha.find_template_companion(template.name)
        template_companion = list(template_companion)

    if not extensions and not no_extension_file:
        for file in template_companion:
            if file.endswith(yasha.EXTENSION_FILE_FORMATS):
                extensions = click.open_file(file, "rb")
                break

    if extensions:
        load_extensions(extensions)

    if not variables and not no_variable_file:
        for file in template_companion:
            if file.endswith(tuple(PARSERS.keys())):
                variables = click.open_file(file, "rb")
                break

    if not output:
        if template.name == "<stdin>":
            output = click.open_file("-", "wb")
        else:
            output = os.path.splitext(template.name)[0]
            output = click.open_file(output, "wb", lazy=True)

    if m or md:
        deps = [os.path.relpath(template.name)]
        if variables:
            deps.append(os.path.relpath(variables.name))
        if extensions:
            deps.append(os.path.relpath(extensions.name))
        for d in yasha.find_referenced_templates(template, include_path):
            deps.append(os.path.relpath(d))

        deps = os.path.relpath(output.name) + ": " + " ".join(deps)
        if m:
            click.echo(deps)
            return  # Template won't be rendered
        if md:
            deps += os.linesep
            output_d = click.open_file(output.name + ".d", "wb")
            output_d.write(deps.encode(yasha.ENCODING))

    # Load Jinja
    jinja = yasha.load_jinja(
        path=include_path,
        tests=TESTS,
        filters=FILTERS,
        classes=CLASSES,
        trim_blocks=not no_trim_blocks,
        lstrip_blocks=not no_lstrip_blocks,
        keep_trailing_newline=keep_trailing_newline
    )

    # Get template
    if template.name == "<stdin>":
        stdin = template.read()
        t = jinja.from_string(stdin.decode(yasha.ENCODING))
    else:
        t = jinja.get_template(os.path.basename(template.name))

    # Parse variables
    context = parse_variable_file(variables)
    context.update(yasha.parse_cli_variables(template_variables))

    # Finally render template and save it
    t_stream = t.stream(context)
    t_stream.enable_buffering(size=5)
    t_stream.dump(output, encoding=yasha.ENCODING)



"""
The MIT License (MIT)

Copyright (c) 2015-2017 Kim Blomqvist

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import os
import ast

__version__ = "4.1"

ENCODING = 'utf-8'
EXTENSION_FILE_FORMATS = ('.py', '.yasha', '.j2ext', '.jinja-ext')

def find_template_companion(template, extension='', check=True):
    """
    Returns the first found template companion file
    """

    if check and not os.path.isfile(template):
        yield ''
        return # May be '<stdin>' (click)

    template = os.path.abspath(template)
    template_dirname = os.path.dirname(template)
    template_basename = os.path.basename(template).split('.')

    current_path = template_dirname
    stop_path = os.path.commonprefix((os.getcwd(), current_path))
    stop_path = os.path.dirname(stop_path)

    token = template_basename[0] + '.'

    while True:

        for file in sorted(os.listdir(current_path)):
            if not file.startswith(token):
                continue
            if not file.endswith(extension):
                continue

            file_parts = file.split('.')
            for i in range(1, len(template_basename)):
                if template_basename[:-i] != file_parts[:-1]:
                    continue
                if current_path == template_dirname:
                    if file_parts == template_basename:
                        continue # Do not accept template itself

                yield os.path.join(current_path, file)

        if current_path == stop_path:
            break

        # cd ..
        current_path = os.path.split(current_path)[0]


def find_referenced_templates(template, search_path):
    """
    Returns a list of files which can be either {% imported %},
    {% extended %} or {% included %} within a template.
    """
    from jinja2 import Environment, meta
    env = Environment()
    ast = env.parse(template.read())
    referenced_templates = list(meta.find_referenced_templates(ast))

    def realpath(tpl):
        for path in search_path:
            t = os.path.realpath(os.path.join(path, tpl))
            if os.path.isfile(t):
                return t
        return None

    return [realpath(t) for t in referenced_templates if t is not None]


def parse_cli_variables(args):
    variables = dict()
    for i, arg in enumerate(args):
        if arg[:2] != '--':
            continue
        if '=' in arg:
            opt, val = arg[2:].split('=', 1)
        else:
            try:
                if args[i+1] != '--':
                    opt = arg[2:]
                    val = args[i+1]
            except IndexError:
                break
        try:
            val = ast.literal_eval(val)
        except ValueError:
            pass
        except SyntaxError:
            pass
        if isinstance(val, str) and ',' in val:
            # Convert foo,bar,baz to list ['foo', 'bar', 'baz']
            val = val.split(',')
        variables[opt] = val
    return variables


def load_jinja(path, tests, filters, classes, trim_blocks, lstrip_blocks, keep_trailing_newline):
    from jinja2 import Environment, FileSystemLoader
    jinja = Environment(
        loader=FileSystemLoader(path),
        extensions=classes,
        trim_blocks=trim_blocks,
        lstrip_blocks=lstrip_blocks,
        keep_trailing_newline=keep_trailing_newline
    )
    jinja.tests.update(tests)
    jinja.filters.update(filters)
    return jinja



"""
The MIT License (MIT)

Copyright (c) 2015-2017 Kim Blomqvist

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from xml.etree import ElementTree

class SVDFile():
    """SVD File: Entry class to parse CMSIS-SVD file

    SVD = System View Description format
    CMSIS = Cortex Microcontroller Software Interface Standard
    Read more from http://www.keil.com/pack/doc/CMSIS/SVD/html/
    """

    def __init__(self, file):
        if isinstance(file, str):
            self.root = ElementTree.fromstring(file)
        else:
            tree = ElementTree.parse(file)
            self.root = tree.getroot()

        self.cpu = None
        self.device = None
        self.peripherals = []
        self.peripherals_dict = {}  # Lookup by peripheral name

    def parse(self):
        self.cpu = SvdCpu(self.root.find("cpu"))
        self.device = SvdDevice(self.root)

        derived_periphs = []
        for elem in self.root.iter("peripheral"):
            periph = SvdPeripheral(elem, self.device)
            if periph.derivedFrom is not None:
                derived_periphs.append(periph.name)
            self.peripherals.append(periph)
            self.peripherals_dict[periph.name] = periph

        for periph in [self.peripherals_dict[name] for name in derived_periphs]:
            base = self.peripherals_dict[periph.derivedFrom]
            periph.inherit_from(base)


class SvdElement(object):
    props = []
    props_to_integer = []
    props_to_boolean = []

    def __init__(self, element=None, defaults={}, parent=None):
        if element is not None:
            self.from_element(element, defaults)
        if parent is not None:
            self.parent = parent

    def from_element(self, element, defaults={}):
        """Populate object variables from SVD element"""
        if isinstance(defaults, SvdElement):
            defaults = vars(defaults)
        for key in self.props:
            try:
                value = element.find(key).text
            except AttributeError:  # Maybe it's attribute?
                default = defaults[key] if key in defaults else None
                value = element.get(key, default)
            if value is not None:
                if key in self.props_to_integer:
                    try:
                        value = int(value)
                    except ValueError:  # It has to be hex
                        value = int(value, 16)
                elif key in self.props_to_boolean:
                    value = value.lower() in ("yes", "true", "t", "1")

            setattr(self, key, value)

    def inherit_from(self, element):
        for key, value in vars(self).items():
            if not value and key in vars(element):
                value = getattr(element, key)
                setattr(self, key, value)

    def copy(self):
        import copy
        return copy.copy(self)

    def __str__(self):
        from pprint import pformat
        return pformat(vars(self), indent=0)


class SvdDevice(SvdElement):
    """SVD Devices element"""
    props = [
        "schemaVersion", "vendor", "vendorID", "name", "series", "version",
        "description", "licenseText", "headerSystemFilename",
        "headerDefinitionsPrefix", "addressUnitBits", "width", "size",
        "access", "protection", "resetValue", "resetMask"
    ]
    props_to_integer = [
        "width", "size", "resetValue", "resetMask", "addressUnitBits"
    ]


class SvdCpu(SvdElement):
    """SVD CPU section"""
    props = [
        "name", "revision", "endian", "mpuPresent", "fpuPresent", "fpuDP",
        "icachePresent", "dcachePresent", "itcmPresent", "dtcmPresent",
        "vtorPresent", "nvicPrioBits", "vendorSystickConfig",
        "deviceNumInterrupts"
    ]
    props_to_boolean = [
        "mpuPresent", "fpuPresent", "fpuDP", "icachePresent", "dcachePresent",
        "itcmPresent", "dtcmPresent", "vtorPresent"
    ]

class SvdPeripheral(SvdElement):
    """SVD Peripherals Level

    A peripheral is a named collection of registers. A peripheral is mapped
    to a defined base address within the device's address space. A peripheral
    allocates one or more exclusive address blocks relative to its base
    address, such that all described registers fit into the allocated address
    blocks. Allocated addresses without an associated register description
    are automatically considered reserved. The peripheral can be assigned to
    a group of peripherals and may be associated with one or more interrupts.
    """
    props = [
        "derivedFrom", "dim", "dimIncrement", "dimIndex", "dimName", "name",
        "version", "description", "alternatePeripheral", "groupName",
        "appendToName", "headerStructName", "disableCondition", "baseAddress",
        "size", "access", "protection", "resetValue", "resetMask"
    ]
    props_to_integer = [
        "dim", "dimIncrement", "baseAddress", "size", "resetValue",
        "resetMask"
    ]

    def from_element(self, element, defaults={}):
        SvdElement.from_element(self, element, defaults)
        self.registers = []
        self.interrupts = []
        self.addressBlock = None

        try:
            for reg in element.find("registers"):
                if reg.tag == "cluster":
                    self.registers.append(Cluster(reg, self, parent=self))
                elif reg.tag == "register":
                    reg = SvdRegister(reg, self, parent=self)
                    self.registers.extend(reg.fold())
        except TypeError:  # element.findall() may return None
            pass

        try:
            for i in element.findall("interrupt"):
                self.interrupts.append(SvdInterrupt(i, parent=self))
        except TypeError:  # element.findall() may return None
            pass

        try:
            block = element.find("addressBlock")
            self.addressBlock = SvdAddressBlock(block, parent=self)
        except TypeError:
            pass


class SvdRegister(SvdElement):
    """SVD Registers Level

    A register is a named, programmable resource that belongs to a
    peripheral. Registers are mapped to a defined address in the address
    space of the device. An address is specified relative to the peripheral
    base address. The description of a register documents the purpose and
    function of the resource. A debugger requires information about the
    permitted access to a resource as well as side effects triggered by
    read and write accesses respectively.
    """
    props = [
        "derivedFrom", "dim", "dimIncrement", "dimIndex", "dimName",
        "name", "displayName", "description", "alternateGroup",
        "alternateRegister", "addressOffset", "size", "access", "protection",
        "resetValue", "resetMask", "dataType", "modifiedWriteValues",
        "readAction"
    ]
    props_to_integer = [
        "dim", "dimIncrement", "addressOffset", "size", "resetValue",
        "resetMask"
    ]

    def from_element(self, element, defaults={}):
        SvdElement.from_element(self, element, defaults)
        self.fields = []

        if self.dim is not None:
            try:
                self.dimIndex = int(self.dimIndex)
            except ValueError:
                try:
                    start, stop = self.dimIndex.split("-")
                    start, stop = (int(start), int(stop)+1)
                    self.dimIndex = list(range(start, stop))
                except ValueError:
                    self.dimIndex = self.dimIndex.split(",")

        try:
            for elem in element.find("fields"):
                field = SvdField(elem, self, parent=self)
                self.fields.append(field)
        except TypeError:  # element.findall() may return None
            pass

        try:
            elem = element.find("writeConstraint")
            self.writeConstraint = SvdWriteConstraint(elem, parent=self)
        except TypeError:
            pass


    def fold(self):
        """Folds the Register in accordance with it's dimensions.

        If the register is dimensionless, the returned list just
        contains the register itself unchanged. In case the register
        name looks like a C array, the returned list contains the register
        itself, where nothing else than the '%s' placeholder in it's name
        has been replaced with value of the dim element.
        """
        if self.dim is None:
            return [self]
        if self.name.endswith("[%s]"):  # C array like
            self.name = self.name.replace("%s", str(self.dim))
            return [self]

        registers = []
        for offset, index in enumerate(self.dimIndex):
            reg = self.copy()
            reg.name = self.name.replace("%s", str(index))
            reg.addressOffset += offset * reg.dimIncrement

            reg.fields = [field.copy() for field in reg.fields]
            for field in reg.fields:
                field.parent = reg

            reg.dim = reg.dimIndex = reg.dimIncrement = None  # Dimensionless
            registers.append(reg)

        return registers


class Cluster(SvdElement):
    """SVD Cluster extension level

    Cluster adds an optional sub-level within the CMSIS SVD registers level.
    A cluster describes a sequence of neighboring registers within
    a peripheral.
    """
    props = [
        "registers", "derivedFrom", "dim", "dimIncrement", "dimIndex", "name",
        "description", "alternateCluster", "headerStructName", "addressOffset"
    ]
    props_to_integer = ["addressOffset", "dim", "dimIncrement"]

    def from_element(self, element, defaults={}):
        SvdElement.from_element(self, element, {})
        self.registers = []

        # TODO: Should work like Register.to_array(), if there's self.dim
        self.name = self.name.replace("%s", str(self.dim))

        try:
            for elem in element.findall("*"):
                if elem.tag == "cluster":  # Cluster may include yet another cluster
                    self.registers.append(Cluster(elem, defaults, parent=self))
                elif elem.tag == "register":
                    reg = SvdRegister(elem, defaults, parent=self)
                    self.registers.extend(reg.fold())
        except TypeError:  # element.findall() may return None
            pass


class SvdField(SvdElement):
    """SVD Fields level

    All fields of a register are enclosed between the <fields>
    opening and closing tags.
    """
    props = [
        "derivedFrom", "name", "description", "bitOffset", "bitWidth",
        "lsb", "msb", "bitRange", "access", "modifiedWriteValues",
        "readAction"
    ]
    props_to_integer = ["bitOffset", "bitWidth", "lsb", "msb"]

    def from_element(self, element, defaults={}):
        SvdElement.from_element(self, element, defaults)
        self.enumeratedValues = {
            "read": [],
            "write": [],
            "read-write": [],
        }

        if self.bitRange is not None:
            self.msb, self.lsb = self.bitRange[1:-1].split(":")
            self.msb = int(self.msb)
            self.lsb = int(self.lsb)
            self.bitOffset = self.lsb
            self.bitWidth = self.msb - self.lsb + 1
        elif self.bitOffset is not None and self.bitWidth is not None:
            self.lsb = self.bitOffset
            self.msb = self.bitWidth + self.lsb
        self.bitRange = (self.msb, self.lsb)

        try:
            for e in element.findall("enumeratedValues"):
                try:
                    usage = e.find("usage").text
                except AttributeError:
                    usage = "read-write"
                for e in e.findall("enumeratedValue"):
                    enum = SvdEnumeratedValue(e, {}, parent=self)
                    self.enumeratedValues[usage].append(enum)
        except TypeError:  # element.findall() may return None
            pass

        try:
            elem = element.find("writeConstraint")
            self.writeConstraint = SvdWriteConstraint(elem, parent=self)
        except TypeError:
            pass


class SvdEnumeratedValue(SvdElement):
    """SVD Enumerated values Level

    The concept of enumerated values creates a map between unsigned
    integers and an identifier string.
    """
    props = ["derivedFrom", "name", "description", "value", "isDefault"]
    props_to_integer = ["value"]


class SvdInterrupt(SvdElement):
    props = ["name", "description", "value"]
    props_to_integer = ["value"]


class SvdAddressBlock(SvdElement):
    props = ["addressBlock", "offset", "size", "usage", "protection"]
    props_to_integer = ["offset", "size"]


class SvdWriteConstraint(SvdElement):
    props = ["writeAsRead", "useEnumeratedValues"]
    props_to_boolean = ["writeAsRead", "useEnumeratedValues"]

    def from_element(self, element, defaults={}):
        SvdElement.from_element(self, element, defaults)
        try:
            elem = element.find("range")
            minimum = elem.find("minimum").text
            maximum = elem.find("maximum").text
            self.range = (int(minimum), int(maximum))
        except:  # No range
            pass
